# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Some useful math functions."""

from typing import Optional, Tuple, Union

import jax
from jax import numpy as jp


def norm(
    x: jax.Array, axis: Optional[Union[Tuple[int, ...], int]] = None
) -> jax.Array:
  """Calculates a linalg.norm(x) that's safe for gradients at x=0.

  Avoids a poorly defined gradient for jnp.linal.norm(0) see
  https://github.com/google/jax/issues/3058 for details
  Args:
    x: A jnp.array
    axis: The axis along which to compute the norm

  Returns:
    Norm of the array x.
  """

  is_zero = jp.allclose(x, 0.0)
  # temporarily swap x with ones if is_zero, then swap back
  x = jp.where(is_zero, jp.ones_like(x), x)
  n = jp.linalg.norm(x, axis=axis)
  n = jp.where(is_zero, 0.0, n)
  return n


def normalize_with_norm(
    x: jax.Array, axis: Optional[Union[Tuple[int, ...], int]] = None
) -> Tuple[jax.Array, jax.Array]:
  """Normalizes an array.

  Args:
    x: A jnp.array
    axis: The axis along which to compute the norm

  Returns:
    A tuple of (normalized array x, the norm).
  """
  n = norm(x, axis=axis)
  x = x / (n + 1e-6 * (n == 0.0))
  return x, n


def normalize(
    x: jax.Array, axis: Optional[Union[Tuple[int, ...], int]] = None
) -> jax.Array:
  """Normalizes an array.

  Args:
    x: A jnp.array
    axis: The axis along which to compute the norm

  Returns:
    normalized array x
  """
  return normalize_with_norm(x, axis=axis)[0]


def rotate(vec: jax.Array, quat: jax.Array) -> jax.Array:
  """Rotates a vector vec by a unit quaternion quat.

  Args:
    vec: (3,) a vector
    quat: (4,) a quaternion

  Returns:
    ndarray(3) containing vec rotated by quat.
  """
  if len(vec.shape) != 1:
    raise ValueError('vec must have no batch dimensions.')
  s, u = quat[0], quat[1:]
  r = 2 * (jp.dot(u, vec) * u) + (s * s - jp.dot(u, u)) * vec
  r = r + 2 * s * jp.cross(u, vec)
  return r


def quat_inv(q: jp.ndarray) -> jp.ndarray:
  """Calculates the inverse of quaternion q.

  Args:
    q: (4,) quaternion [w, x, y, z]

  Returns:
    The inverse of q, where qmult(q, inv_quat(q)) = [1, 0, 0, 0].
  """
  return q * jp.array([1, -1, -1, -1])


def quat_sub(u: jax.Array, v: jax.Array) -> jax.Array:
  """Subtracts two quaternions (u - v) as a 3D velocity."""
  q = quat_mul(quat_inv(v), u)
  axis, angle = quat_to_axis_angle(q)
  return axis * angle


def quat_mul(u: jax.Array, v: jax.Array) -> jax.Array:
  """Multiplies two quaternions.

  Args:
    u: (4,) quaternion (w,x,y,z)
    v: (4,) quaternion (w,x,y,z)

  Returns:
    A quaternion u * v.
  """
  return jp.array([
      u[0] * v[0] - u[1] * v[1] - u[2] * v[2] - u[3] * v[3],
      u[0] * v[1] + u[1] * v[0] + u[2] * v[3] - u[3] * v[2],
      u[0] * v[2] - u[1] * v[3] + u[2] * v[0] + u[3] * v[1],
      u[0] * v[3] + u[1] * v[2] - u[2] * v[1] + u[3] * v[0],
  ])


def quat_mul_axis(q: jax.Array, axis: jax.Array) -> jax.Array:
  """Multiplies a quaternion and an axis.

  Args:
    q: (4,) quaternion (w,x,y,z)
    axis: (3,) axis (x,y,z)

  Returns:
    A quaternion q * axis
  """
  return jp.array([
      -q[1] * axis[0] - q[2] * axis[1] - q[3] * axis[2],
      q[0] * axis[0] + q[2] * axis[2] - q[3] * axis[1],
      q[0] * axis[1] + q[3] * axis[0] - q[1] * axis[2],
      q[0] * axis[2] + q[1] * axis[1] - q[2] * axis[0],
  ])


# TODO(erikfrey): benchmark this against brax's quat_to_3x3
def quat_to_mat(q: jax.Array) -> jax.Array:
  """Converts a quaternion into a 9-dimensional rotation matrix."""
  q = jp.outer(q, q)

  return jp.array([
      [
          q[0, 0] + q[1, 1] - q[2, 2] - q[3, 3],
          2 * (q[1, 2] - q[0, 3]),
          2 * (q[1, 3] + q[0, 2]),
      ],
      [
          2 * (q[1, 2] + q[0, 3]),
          q[0, 0] - q[1, 1] + q[2, 2] - q[3, 3],
          2 * (q[2, 3] - q[0, 1]),
      ],
      [
          2 * (q[1, 3] - q[0, 2]),
          2 * (q[2, 3] + q[0, 1]),
          q[0, 0] - q[1, 1] - q[2, 2] + q[3, 3],
      ],
  ])


def quat_to_axis_angle(q: jax.Array) -> Tuple[jax.Array, jax.Array]:
  """Converts a quaternion into axis and angle."""
  axis, sin_a_2 = normalize_with_norm(q[1:])
  angle = 2 * jp.arctan2(sin_a_2, q[0])
  angle = jp.where(angle > jp.pi, angle - 2 * jp.pi, angle)

  return axis, angle


def axis_angle_to_quat(axis: jax.Array, angle: jax.Array) -> jax.Array:
  """Provides a quaternion that describes rotating around axis by angle.

  Args:
    axis: (3,) axis (x,y,z)
    angle: () float angle to rotate by

  Returns:
    A quaternion that rotates around axis by angle
  """
  s, c = jp.sin(angle * 0.5), jp.cos(angle * 0.5)
  return jp.insert(axis * s, 0, c)


def quat_integrate(q: jax.Array, v: jax.Array, dt: jax.Array) -> jax.Array:
  """Integrates a quaternion given angular velocity and dt."""
  v, norm_ = normalize_with_norm(v)
  angle = dt * norm_
  q_res = axis_angle_to_quat(v, angle)
  q_res = quat_mul(q, q_res)
  return normalize(q_res)


def inert_mul(i: jax.Array, v: jax.Array) -> jax.Array:
  """Multiply inertia by motion, producing force.

  Args:
    i: (10,) inertia (inertia matrix, position, mass)
    v: (6,) spatial motion

  Returns:
    resultant force
  """
  tri_id = jp.array([[0, 3, 4], [3, 1, 5], [4, 5, 2]])  # cinert inr order
  inr, pos, mass = i[tri_id], i[6:9], i[9]
  ang = jp.dot(inr, v[:3]) + jp.cross(pos, v[3:])
  vel = mass * v[3:] - jp.cross(pos, v[:3])
  return jp.concatenate((ang, vel))


def transform_motion(vel: jax.Array, offset: jax.Array, rotmat: jax.Array):
  """Transform spatial motion.

  Args:
    vel: (6,) spatial motion (3 angular, 3 linear)
    offset: (3,) translation
    rotmat: (3, 3) rotation

  Returns:
    6d spatial velocity
  """
  # TODO(robotics-simulation): are quaternions faster here
  ang, vel = vel[:3], vel[3:]
  vel = rotmat.T @ (vel - jp.cross(offset, ang))
  ang = rotmat.T @ ang
  return jp.concatenate([ang, vel])


def motion_cross(u, v):
  """Cross product of two motions.

  Args:
    u: (6,) spatial motion
    v: (6,) spatial motion

  Returns:
    resultant spatial motion
  """
  ang = jp.cross(u[:3], v[:3])
  vel = jp.cross(u[3:], v[:3]) + jp.cross(u[:3], v[3:])
  return jp.concatenate((ang, vel))


def motion_cross_force(v, f):
  """Cross product of a motion and force.

  Args:
    v: (6,) spatial motion
    f: (6,) force

  Returns:
    resultant force
  """
  ang = jp.cross(v[:3], f[:3]) + jp.cross(v[3:], f[3:])
  vel = jp.cross(v[:3], f[3:])
  return jp.concatenate((ang, vel))


def orthogonals(a: jax.Array) -> Tuple[jax.Array, jax.Array]:
  """Returns orthogonal vectors `b` and `c`, given a vector `a`."""
  y, z = jp.array([0, 1, 0]), jp.array([0, 0, 1])
  b = jp.where((-0.5 < a[1]) & (a[1] < 0.5), y, z)
  b = b - a * a.dot(b)
  # normalize b. however if a is a zero vector, zero b as well.
  b = normalize(b) * jp.any(a)
  return b, jp.cross(a, b)


def make_frame(a: jax.Array) -> jax.Array:
  """Makes a right-handed 3D frame given a direction."""
  a = normalize(a)
  b, c = orthogonals(a)
  return jp.array([a, b, c])


# Geometry.


def closest_segment_point(
    a: jax.Array, b: jax.Array, pt: jax.Array
) -> jax.Array:
  """Returns the closest point on the a-b line segment to a point pt."""
  ab = b - a
  t = jp.dot(pt - a, ab) / (jp.dot(ab, ab) + 1e-6)
  return a + jp.clip(t, 0.0, 1.0) * ab


def closest_segment_point_and_dist(
    a: jax.Array, b: jax.Array, pt: jax.Array
) -> Tuple[jax.Array, jax.Array]:
  """Returns closest point on the line segment and the distance squared."""
  closest = closest_segment_point(a, b, pt)
  dist = (pt - closest).dot(pt - closest)
  return closest, dist


def closest_segment_to_segment_points(
    a0: jax.Array, a1: jax.Array, b0: jax.Array, b1: jax.Array
) -> Tuple[jax.Array, jax.Array]:
  """Returns closest points between two line segments."""
  # Gets the closest segment points by first finding the closest points
  # between two lines. Points are then clipped to be on the line segments
  # and edge cases with clipping are handled.
  dir_a, len_a = normalize_with_norm(a1 - a0)
  dir_b, len_b = normalize_with_norm(b1 - b0)

  # Segment mid-points.
  half_len_a = len_a * 0.5
  half_len_b = len_b * 0.5
  a_mid = a0 + dir_a * half_len_a
  b_mid = b0 + dir_b * half_len_b

  # Translation between two segment mid-points.
  trans = a_mid - b_mid

  # Parametrize points on each line as follows:
  #  point_on_a = a_mid + t_a * dir_a
  #  point_on_b = b_mid + t_b * dir_b
  # and analytically minimize the distance between the two points.
  dira_dot_dirb = dir_a.dot(dir_b)
  dira_dot_trans = dir_a.dot(trans)
  dirb_dot_trans = dir_b.dot(trans)
  denom = 1 - dira_dot_dirb * dira_dot_dirb

  orig_t_a = (-dira_dot_trans + dira_dot_dirb * dirb_dot_trans) / (denom + 1e-6)
  orig_t_b = dirb_dot_trans + orig_t_a * dira_dot_dirb
  t_a = jp.clip(orig_t_a, -half_len_a, half_len_a)
  t_b = jp.clip(orig_t_b, -half_len_b, half_len_b)

  best_a = a_mid + dir_a * t_a
  best_b = b_mid + dir_b * t_b

  # Resolve edge cases where both closest points are clipped to the segment
  # endpoints by recalculating the closest segment points for the current
  # clipped points, and then picking the pair of points with smallest
  # distance. An example of this edge case is when lines intersect but line
  # segments don't.
  new_a, d1 = closest_segment_point_and_dist(a0, a1, best_b)
  new_b, d2 = closest_segment_point_and_dist(b0, b1, best_a)
  best_a = jp.where(d1 < d2, new_a, best_a)
  best_b = jp.where(d1 < d2, best_b, new_b)

  return best_a, best_b



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for mesh.py."""

from absl.testing import absltest
from mujoco.mjx._src import mesh
import numpy as np


class GeomMeshKwargsTest(absltest.TestCase):

  def test_pyramid(self):
    """Tests that a triangulated pyramid converts to merged coplanar faces."""
    vert = np.array([
        [-0.025, 0.05, 0.05],
        [-0.025, -0.05, -0.05],
        [-0.025, -0.05, 0.05],
        [-0.025, 0.05, -0.05],
        [0.075, 0.0, 0.0],
    ])
    face = np.array(
        [[0, 1, 2], [0, 3, 1], [0, 4, 3], [0, 2, 4], [2, 1, 4], [1, 3, 4]]
    )
    h = mesh._geom_mesh_kwargs(vert, face)

    # get index of vertices in h['geom_convex_vert'] for vertices in vert
    dist = np.repeat(vert, vert.shape[0], axis=0) - np.tile(
        h['geom_convex_vert'], (vert.shape[0], 1)
    )
    dist = (dist**2).sum(axis=1).reshape((vert.shape[0], -1))
    vidx = np.argmin(dist, axis=0)

    # check verts
    np.testing.assert_array_equal(h['geom_convex_vert'], vert[vidx])

    # check face vertices
    map_ = {v: k for k, v in enumerate(vidx)}
    h_face = np.vectorize(map_.get)(h['geom_convex_face'])
    face_verts = sorted([tuple(sorted(set(s))) for s in h_face.tolist()])
    expected_face_verts = sorted([
        (0, 3, 4), (1, 3, 4), (0, 2, 4), (0, 1, 2, 3), (1, 2, 4)])
    self.assertSequenceEqual(
        face_verts,
        expected_face_verts,
    )

    # check edges
    unique_edge = np.vectorize(map_.get)(h['geom_convex_edge'])
    unique_edge = np.array(sorted(unique_edge.tolist()))
    np.testing.assert_array_equal(
        unique_edge,
        np.array([[0, 2], [0, 3], [0, 4], [1, 4], [2, 4], [3, 4]]),
    )

    # face normals
    self.assertEqual(h['geom_convex_facenormal'].shape, (5, 3))


class ConvexHull2DTest(absltest.TestCase):

  def test_convex_hull_2d_axis1(self):
    """Tests for the correct winding order of a polgyon with +y normal."""
    pts = np.array([
        [-0.04634297, -0.06652775, 0.05853534],
        [-0.01877651, -0.08309858, -0.05236476],
        [0.02362804, -0.08010745, 0.05499557],
        [0.04066505, -0.09034877, -0.01354446],
        [-0.07255043, -0.06837638, -0.00781699],
    ])
    normal = np.array([-0.18467607, -0.97768016, 0.10018111])
    idx = mesh._convex_hull_2d(pts, normal)
    expected = np.cross(pts[idx][1] - pts[idx][0], pts[idx][2] - pts[idx][0])
    expected /= np.linalg.norm(expected)
    np.testing.assert_array_almost_equal(normal, expected)

  def test_convex_hull_2d_axis2(self):
    """Tests for the correct winding order for a polgyon with +z normal."""
    pts = np.array([
        [0.08607829, -0.03881998, -0.03291714],
        [-0.01877651, -0.08309858, -0.05236476],
        [0.05470364, 0.00027677, -0.08371042],
        [-0.01010019, -0.02708892, -0.0957297],
        [0.04066505, -0.09034877, -0.01354446],
    ])
    normal = np.array([0.3839915, -0.60171936, -0.70034587])
    idx = mesh._convex_hull_2d(pts, normal)
    expected = np.cross(pts[idx][1] - pts[idx][0], pts[idx][2] - pts[idx][0])
    expected /= np.linalg.norm(expected)
    np.testing.assert_array_almost_equal(normal, expected)


class UniqueEdgesTest(absltest.TestCase):

  def test_tetrahedron_edges(self):
    """Tests unique edges for a tetrahedron."""
    vert = np.array(
        [[-0.1, 0.0, -0.1], [0.0, 0.1, 0.1], [0.1, 0.0, -0.1], [0.0, -0.1, 0.1]]
    )
    face = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 1], [2, 1, 3]])
    idx = mesh._get_unique_edges(vert, face)
    np.testing.assert_array_equal(
        idx, np.array([[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]])
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper that automatically registers dataclass as a Jax PyTree."""

import copy
import dataclasses

import typing
from typing import Any, Dict, Optional, Sequence, TypeVar
import jax
import numpy as np

_T = TypeVar('_T')


def dataclass(clz: _T) -> _T:
  """Wraps a dataclass with metadata for which fields are pytrees.

  This is based off flax.struct.dataclass, but instead of using field
  descriptors to specify which fields are pytrees, we follow a simple rule:
  a leaf field is a pytree node if and only if it's a jax.Array

  Args:
    clz: the class to register as a dataclass

  Returns:
    the resulting dataclass, registered with Jax
  """
  data_clz = dataclasses.dataclass(frozen=True)(clz)
  meta_fields, data_fields = [], []
  for field in dataclasses.fields(data_clz):
    if any((
        field.type is jax.Array,
        dataclasses.is_dataclass(field.type),
        jax.Array in typing.get_args(field.type),
    )):
      data_fields.append(field)
    else:
      meta_fields.append(field)

  def replace(self, **updates):
    """"Returns a new object replacing the specified fields with new values."""
    return dataclasses.replace(self, **updates)

  data_clz.replace = replace

  def iterate_clz_with_keys(x):
    # numpy arrays are not hashable, so convert them to tuples for jit cache
    to_tup = lambda x: tuple(x) if len(x.shape) == 1 else tuple(map(to_tup, x))

    def to_meta(field, obj):
      val = getattr(obj, field.name)
      return to_tup(val) if isinstance(val, np.ndarray) else val

    def to_data(field, obj):
      return (jax.tree_util.GetAttrKey(field.name), getattr(obj, field.name))

    data = tuple(to_data(f, x) for f in data_fields)
    meta = tuple(to_meta(f, x) for f in meta_fields)
    return data, meta

  def clz_from_iterable(meta, data):

    def from_meta(field, meta):
      if field.type is np.ndarray:
        return (field.name, np.array(meta))
      else:
        return (field.name, meta)

    from_data = lambda field, meta: (field.name, meta)

    meta_args = tuple(from_meta(f, m) for f, m in zip(meta_fields, meta))
    data_args = tuple(from_data(f, m) for f, m in zip(data_fields, data))

    return data_clz(**dict(meta_args + data_args))

  jax.tree_util.register_pytree_with_keys(
      data_clz, iterate_clz_with_keys, clz_from_iterable
  )

  return data_clz


TNode = TypeVar('TNode', bound='PyTreeNode')


class PyTreeNode:
  """Base class for dataclasses that should act like a JAX pytree node.

  This base class additionally avoids type checking errors when using PyType.
  """

  def __init_subclass__(cls):
    dataclass(cls)

  def __init__(self, *args, **kwargs):
    # stub for pytype
    raise NotImplementedError

  def replace(self: TNode, **overrides) -> TNode:
    # stub for pytype
    raise NotImplementedError

  @classmethod
  def fields(cls) -> tuple[dataclasses.Field[Any], ...]:
    return dataclasses.fields(cls)

  def tree_replace(
      self, params: Dict[str, Optional[jax.typing.ArrayLike]]
  ) -> 'PyTreeNode':
    new = self
    for k, v in params.items():
      new = _tree_replace(new, k.split('.'), v)
    return new


def _tree_replace(
    base: PyTreeNode,
    attr: Sequence[str],
    val: Optional[jax.typing.ArrayLike],
) -> PyTreeNode:
  """Sets attributes in a struct.dataclass with values."""
  if not attr:
    return base

  # special case for List attribute
  if len(attr) > 1 and isinstance(getattr(base, attr[0]), list):
    lst = copy.deepcopy(getattr(base, attr[0]))

    for i, g in enumerate(lst):
      if not hasattr(g, attr[1]):
        continue
      v = val if not hasattr(val, '__iter__') else val[i]
      lst[i] = _tree_replace(g, attr[1:], v)

    return base.replace(**{attr[0]: lst})

  if len(attr) == 1:
    return base.replace(**{attr[0]: val})

  return base.replace(
      **{attr[0]: _tree_replace(getattr(base, attr[0]), attr[1:], val)}
  )



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Collide geometries."""

from typing import Callable, Dict, Optional, Sequence, Tuple, Union

import jax
from jax import numpy as jp
import mujoco
from mujoco.mjx._src import collision_base
# pylint: disable=g-importing-member
from mujoco.mjx._src.collision_base import Candidate
from mujoco.mjx._src.collision_base import CandidateSet
from mujoco.mjx._src.collision_base import GeomInfo
from mujoco.mjx._src.collision_base import SolverParams
from mujoco.mjx._src.collision_convex import capsule_convex
from mujoco.mjx._src.collision_convex import convex_convex
from mujoco.mjx._src.collision_convex import plane_convex
from mujoco.mjx._src.collision_convex import sphere_convex
from mujoco.mjx._src.collision_primitive import capsule_capsule
from mujoco.mjx._src.collision_primitive import plane_capsule
from mujoco.mjx._src.collision_primitive import plane_sphere
from mujoco.mjx._src.collision_primitive import sphere_capsule
from mujoco.mjx._src.collision_primitive import sphere_sphere
from mujoco.mjx._src.types import Contact
from mujoco.mjx._src.types import Data
from mujoco.mjx._src.types import DisableBit
from mujoco.mjx._src.types import GeomType
from mujoco.mjx._src.types import Model
# pylint: enable=g-importing-member


# pair-wise collision functions
_COLLISION_FUNC = {
    (GeomType.PLANE, GeomType.SPHERE): plane_sphere,
    (GeomType.PLANE, GeomType.CAPSULE): plane_capsule,
    (GeomType.PLANE, GeomType.BOX): plane_convex,
    (GeomType.PLANE, GeomType.MESH): plane_convex,
    (GeomType.SPHERE, GeomType.SPHERE): sphere_sphere,
    (GeomType.SPHERE, GeomType.CAPSULE): sphere_capsule,
    (GeomType.SPHERE, GeomType.BOX): sphere_convex,
    (GeomType.SPHERE, GeomType.MESH): sphere_convex,
    (GeomType.CAPSULE, GeomType.CAPSULE): capsule_capsule,
    (GeomType.CAPSULE, GeomType.BOX): capsule_convex,
    (GeomType.CAPSULE, GeomType.MESH): capsule_convex,
    (GeomType.BOX, GeomType.BOX): convex_convex,
    (GeomType.BOX, GeomType.MESH): convex_convex,
    (GeomType.MESH, GeomType.MESH): convex_convex,
}


def get_collision_fn(
    key: Tuple[Union[GeomType, mujoco.mjtGeom], Union[GeomType, mujoco.mjtGeom]]
) -> Optional[Callable[[GeomInfo, GeomInfo], collision_base.Contact]]:
  """Returns a collision function given a pair of geom types."""
  return _COLLISION_FUNC.get(key, None)


def _add_candidate(
    result: CandidateSet,
    m: Union[Model, mujoco.MjModel],
    g1: int,
    g2: int,
    ipair: int = -1,
):
  """Adds a candidate to test for collision."""
  t1, t2 = m.geom_type[g1], m.geom_type[g2]
  if t1 > t2:
    t1, t2, g1, g2 = t2, t1, g2, g1

  # MuJoCo does not collide planes with other planes or hfields
  if t1 == GeomType.PLANE and t2 == GeomType.PLANE:
    return
  if t1 == GeomType.PLANE and t2 == GeomType.HFIELD:
    return

  def mesh_key(i):
    convex_data = [[None] * m.ngeom] * 3
    if isinstance(m, Model):
      convex_data = [m.geom_convex_face, m.geom_convex_vert, m.geom_convex_edge]
    key = tuple((-1,) if v[i] is None else v[i].shape for v in convex_data)
    return key

  k1, k2 = mesh_key(g1), mesh_key(g2)

  candidates = {(c.geom1, c.geom2) for c in result.get((t1, t2, k1, k2), [])}
  if (g1, g2) in candidates:
    return

  if ipair > -1:
    candidate = Candidate(g1, g2, ipair, -1, m.pair_dim[ipair])
  elif m.geom_priority[g1] != m.geom_priority[g2]:
    gp = g1 if m.geom_priority[g1] > m.geom_priority[g2] else g2
    candidate = Candidate(g1, g2, -1, gp, m.geom_condim[gp])
  else:
    dim = max(m.geom_condim[g1], m.geom_condim[g2])
    candidate = Candidate(g1, g2, -1, -1, dim)

  result.setdefault((t1, t2, k1, k2), []).append(candidate)


def _pair_params(
    m: Model,
    candidates: Sequence[Candidate],
) -> SolverParams:
  """Gets solver params for pair geoms."""
  ipair = jp.array([c.ipair for c in candidates])
  friction = jp.clip(m.pair_friction[ipair], a_min=mujoco.mjMINMU)
  solref = m.pair_solref[ipair]
  solreffriction = m.pair_solreffriction[ipair]
  solimp = m.pair_solimp[ipair]
  margin = m.pair_margin[ipair]
  gap = m.pair_gap[ipair]

  return SolverParams(friction, solref, solreffriction, solimp, margin, gap)


def _priority_params(
    m: Model,
    candidates: Sequence[Candidate],
) -> SolverParams:
  """Gets solver params from priority geoms."""
  geomp = jp.array([c.geomp for c in candidates])
  friction = m.geom_friction[geomp][:, jp.array([0, 0, 1, 2, 2])]
  solref = m.geom_solref[geomp]
  solreffriction = jp.zeros(geomp.shape + (mujoco.mjNREF,))
  solimp = m.geom_solimp[geomp]
  g = jp.array([(c.geom1, c.geom2) for c in candidates])
  margin = jp.amax(m.geom_margin[g.T], axis=0)
  gap = jp.amax(m.geom_gap[g.T], axis=0)

  return SolverParams(friction, solref, solreffriction, solimp, margin, gap)


def _dynamic_params(
    m: Model,
    candidates: Sequence[Candidate],
) -> SolverParams:
  """Gets solver params for dynamic geoms."""
  g1 = jp.array([c.geom1 for c in candidates])
  g2 = jp.array([c.geom2 for c in candidates])

  friction = jp.maximum(m.geom_friction[g1], m.geom_friction[g2])
  # copy friction terms for the full geom pair
  friction = friction[:, jp.array([0, 0, 1, 2, 2])]

  minval = jp.array(mujoco.mjMINVAL)
  solmix1, solmix2 = m.geom_solmix[g1], m.geom_solmix[g2]
  mix = solmix1 / (solmix1 + solmix2)
  mix = jp.where((solmix1 < minval) & (solmix2 < minval), 0.5, mix)
  mix = jp.where((solmix1 < minval) & (solmix2 >= minval), 0.0, mix)
  mix_fn = jax.vmap(lambda a, b, m: m * a + (1 - m) * b)

  solref1, solref2 = m.geom_solref[g1], m.geom_solref[g2]
  solref = jp.minimum(solref1, solref2)
  s_mix = mix_fn(solref1, solref2, mix)
  solref = jp.where((solref1[0] > 0) & (solref2[0] > 0), s_mix, solref)
  solreffriction = jp.zeros(g1.shape + (mujoco.mjNREF,))
  solimp = mix_fn(m.geom_solimp[g1], m.geom_solimp[g2], mix)
  margin = jp.maximum(m.geom_margin[g1], m.geom_margin[g2])
  gap = jp.maximum(m.geom_gap[g1], m.geom_gap[g2])

  return SolverParams(friction, solref, solreffriction, solimp, margin, gap)


def _pair_info(
    m: Model, d: Data, geom1: Sequence[int], geom2: Sequence[int]
) -> Tuple[GeomInfo, GeomInfo, Sequence[Dict[str, Optional[int]]]]:
  """Returns geom pair info for calculating collision."""
  g1, g2 = jp.array(geom1), jp.array(geom2)
  info1 = GeomInfo(
      d.geom_xpos[g1],
      d.geom_xmat[g1],
      m.geom_size[g1],
  )
  info2 = GeomInfo(
      d.geom_xpos[g2],
      d.geom_xmat[g2],
      m.geom_size[g2],
  )
  in_axes1 = in_axes2 = jax.tree_map(lambda x: 0, info1)
  if m.geom_convex_face[geom1[0]] is not None:
    info1 = info1.replace(
        face=jp.stack([m.geom_convex_face[i] for i in geom1]),
        vert=jp.stack([m.geom_convex_vert[i] for i in geom1]),
        edge=jp.stack([m.geom_convex_edge[i] for i in geom1]),
        facenorm=jp.stack([m.geom_convex_facenormal[i] for i in geom1]),
    )
    in_axes1 = in_axes1.replace(face=0, vert=0, edge=0, facenorm=0)
  if m.geom_convex_face[geom2[0]] is not None:
    info2 = info2.replace(
        face=jp.stack([m.geom_convex_face[i] for i in geom2]),
        vert=jp.stack([m.geom_convex_vert[i] for i in geom2]),
        edge=jp.stack([m.geom_convex_edge[i] for i in geom2]),
        facenorm=jp.stack([m.geom_convex_facenormal[i] for i in geom2]),
    )
    in_axes2 = in_axes2.replace(face=0, vert=0, edge=0, facenorm=0)
  return info1, info2, [in_axes1, in_axes2]


def _body_pair_filter(
    m: Union[Model, mujoco.MjModel], b1: int, b2: int
) -> bool:
  """Filters body pairs for collision."""
  dsbl_filterparent = m.opt.disableflags & DisableBit.FILTERPARENT
  weld1 = m.body_weldid[b1]
  weld2 = m.body_weldid[b2]
  parent_weld1 = m.body_weldid[m.body_parentid[weld1]]
  parent_weld2 = m.body_weldid[m.body_parentid[weld2]]

  if weld1 == weld2:
    # filter out self-collisions
    return True

  if (
      not dsbl_filterparent
      and weld1 != 0
      and weld2 != 0
      and (weld1 == parent_weld2 or weld2 == parent_weld1)
  ):
    # filter out parent-child collisions
    return True

  return False


def _collide_geoms(
    m: Model,
    d: Data,
    geom_types: Tuple[GeomType, GeomType],
    candidates: Sequence[Candidate],
) -> Contact:
  """Collides a geom pair."""
  fn = get_collision_fn(geom_types)
  if not fn:
    return Contact.zero()

  # group sol params by different candidate types
  typ_cands = {}
  for c in candidates:
    typ = (c.ipair > -1, c.geomp > -1)
    typ_cands.setdefault(typ, []).append(c)

  geom1, geom2, params = [], [], []
  for (pair, priority), candidates in typ_cands.items():
    geom1.extend([c.geom1 for c in candidates])
    geom2.extend([c.geom2 for c in candidates])
    if pair:
      params.append(_pair_params(m, candidates))
    elif priority:
      params.append(_priority_params(m, candidates))
    else:
      params.append(_dynamic_params(m, candidates))

  # call contact function
  g1, g2, in_axes = _pair_info(m, d, geom1, geom2)
  res = jax.vmap(fn, in_axes=in_axes)(g1, g2)
  dist, pos, frame = jax.tree_map(jp.concatenate, res)

  params = jax.tree_map(lambda *x: jp.concatenate(x), *params)
  geom1, geom2 = jp.array(geom1), jp.array(geom2)
  # repeat params by the number of contacts per geom pair
  n_repeat = dist.shape[-1] // geom1.shape[0]
  geom1, geom2, params = jax.tree_map(
      lambda x: jp.repeat(x, n_repeat, axis=0),
      (geom1, geom2, params),
  )

  con = Contact(
      dist=dist,
      pos=pos,
      frame=frame,
      includemargin=params.margin - params.gap,
      friction=params.friction,
      solref=params.solref,
      solreffriction=params.solreffriction,
      solimp=params.solimp,
      geom1=geom1,
      geom2=geom2,
  )
  return con


def _max_contact_points(m: Union[Model, mujoco.MjModel]) -> int:
  """Returns the maximum number of contact points when set as a numeric."""
  for i in range(m.nnumeric):
    name = m.names[m.name_numericadr[i] :].decode('utf-8').split('\x00', 1)[0]
    if name == 'max_contact_points':
      return int(m.numeric_data[m.numeric_adr[i]])

  return -1


def collision_candidates(m: Union[Model, mujoco.MjModel]) -> CandidateSet:
  """Returns candidates for collision checking."""
  candidate_set = {}

  for ipair in range(m.npair):
    g1, g2 = m.pair_geom1[ipair], m.pair_geom2[ipair]
    _add_candidate(candidate_set, m, g1, g2, ipair)

  body_pairs = []
  exclude_signature = set(m.exclude_signature)
  for b1 in range(m.nbody):
    for b2 in range(b1, m.nbody):
      signature = (b1 << 16) + (b2)
      if signature in exclude_signature:
        continue
      if _body_pair_filter(m, b1, b2):
        continue
      body_pairs.append((b1, b2))

  for b1, b2 in body_pairs:
    start1 = m.body_geomadr[b1]
    end1 = m.body_geomadr[b1] + m.body_geomnum[b1]
    for g1 in range(start1, end1):
      start2 = m.body_geomadr[b2]
      end2 = m.body_geomadr[b2] + m.body_geomnum[b2]
      for g2 in range(start2, end2):
        mask = m.geom_contype[g1] & m.geom_conaffinity[g2]
        mask |= m.geom_contype[g2] & m.geom_conaffinity[g1]
        if mask != 0:
          _add_candidate(candidate_set, m, g1, g2)

  return candidate_set


def ncon(m: Union[Model, mujoco.MjModel]) -> int:
  """Returns the number of contacts computed in MJX given a model."""
  if m.opt.disableflags & DisableBit.CONTACT:
    return 0

  candidates = collision_candidates(m)
  max_count = _max_contact_points(m)

  count = 0
  for k, v in candidates.items():
    fn = get_collision_fn(k[0:2])
    if fn is None:
      continue
    count += len(v) * fn.ncon  # pytype: disable=attribute-error

  return min(max_count, count) if max_count > -1 else count


def collision(m: Model, d: Data) -> Data:
  """Collides geometries."""
  if ncon(m) == 0:
    return d.replace(contact=Contact.zero())

  candidate_set = collision_candidates(m)

  contacts = []
  for key, candidates in candidate_set.items():
    geom_types = key[0:2]
    contacts.append(_collide_geoms(m, d, geom_types, candidates))

  if not contacts:
    raise RuntimeError('No contacts found.')

  contact = jax.tree_map(lambda *x: jp.concatenate(x), *contacts)

  max_contact_points = _max_contact_points(m)
  if max_contact_points > -1 and contact.dist.shape[0] > max_contact_points:
    # get top-k contacts
    _, idx = jax.lax.top_k(-contact.dist, k=max_contact_points)
    contact = jax.tree_map(lambda x, idx=idx: jp.take(x, idx, axis=0), contact)

  return d.replace(contact=contact)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for smooth dynamics functions."""

from absl.testing import absltest
import jax
from jax import numpy as jp
import mujoco
from mujoco import mjx
from mujoco.mjx._src import test_util
import numpy as np

# tolerance for difference between MuJoCo and MJX smooth calculations - mostly
# due to float precision
_TOLERANCE = 5e-5


def _assert_eq(a, b, name):
  tol = _TOLERANCE * 10  # avoid test noise
  err_msg = f'mismatch: {name}'
  np.testing.assert_allclose(a, b, err_msg=err_msg, atol=tol, rtol=tol)


def _assert_attr_eq(a, b, attr):
  _assert_eq(getattr(a, attr), getattr(b, attr), attr)


class SmoothTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # although we already have generous padding of thresholds, it doesn't hurt
    # to also fix the seed to reduce test flakiness
    np.random.seed(0)

  def test_smooth(self):
    """Tests MJX smooth functions match MuJoCo smooth functions."""

    m = test_util.load_test_file('pendula.xml')
    d = mujoco.MjData(m)
    # give the system a little kick to ensure we have non-identity rotations
    d.qvel = np.random.random(m.nv)
    mujoco.mj_step(m, d, 10)  # let dynamics get state significantly non-zero
    mujoco.mj_forward(m, d)
    mx = mjx.put_model(m)

    # kinematics
    dx = jax.jit(mjx.kinematics)(mx, mjx.put_data(m, d))
    _assert_attr_eq(d, dx, 'xanchor')
    _assert_attr_eq(d, dx, 'xaxis')
    _assert_attr_eq(d, dx, 'xpos')
    _assert_attr_eq(d, dx, 'xquat')
    _assert_eq(d.xmat.reshape((-1, 3, 3)), dx.xmat, 'xmat')
    _assert_attr_eq(d, dx, 'xipos')
    _assert_eq(d.ximat.reshape((-1, 3, 3)), dx.ximat, 'ximat')
    _assert_attr_eq(d, dx, 'geom_xpos')
    _assert_eq(d.geom_xmat.reshape((-1, 3, 3)), dx.geom_xmat, 'geom_xmat')
    _assert_attr_eq(d, dx, 'site_xpos')
    _assert_eq(d.site_xmat.reshape((-1, 3, 3)), dx.site_xmat, 'site_xmat')
    # com_pos
    dx = jax.jit(mjx.com_pos)(mx, mjx.put_data(m, d))
    _assert_attr_eq(d, dx, 'subtree_com')
    _assert_attr_eq(d, dx, 'cinert')
    _assert_attr_eq(d, dx, 'cdof')
    # crb
    dx = jax.jit(mjx.crb)(mx, mjx.put_data(m, d))
    _assert_attr_eq(d, dx, 'crb')
    _assert_attr_eq(d, dx, 'qM')
    # factor_m
    dx = mjx.put_data(m, d)
    dx = jax.jit(mjx.factor_m)(mx, dx, dx.qM)
    _assert_attr_eq(d, dx, 'qLD')
    _assert_attr_eq(d, dx, 'qLDiagInv')
    # com_vel
    dx = jax.jit(mjx.com_vel)(mx, mjx.put_data(m, d))
    _assert_attr_eq(d, dx, 'cvel')
    _assert_attr_eq(d, dx, 'cdof_dot')
    # rne
    dx = jax.jit(mjx.rne)(mx, mjx.put_data(m, d))
    _assert_attr_eq(d, dx, 'qfrc_bias')
    # transmission
    dx = jax.jit(mjx.transmission)(mx, mjx.put_data(m, d))
    _assert_attr_eq(d, dx, 'actuator_length')
    _assert_attr_eq(d, dx, 'actuator_moment')

  def test_mul_m(self):
    m = test_util.load_test_file('pendula.xml')
    d = mujoco.MjData(m)
    # give the system a little kick to ensure we have non-identity rotations
    d.qvel = np.random.random(m.nv)
    mujoco.mj_step(m, d, 10)  # let dynamics get state significantly non-zero
    mujoco.mj_forward(m, d)
    mx = mjx.put_model(m)
    dx = mjx.put_data(m, d)
    vec = np.random.random(m.nv)
    mjx_vec = jax.jit(mjx.mul_m)(mx, dx, jp.array(vec))
    mj_vec = np.zeros(m.nv)
    mujoco.mj_mulM(m, d, mj_vec, vec)
    _assert_eq(mj_vec, mjx_vec, 'mul_m')

  def test_disable_gravity(self):
    m = mujoco.MjModel.from_xml_string("""
        <mujoco>
          <option>
            <flag gravity="disable"/>
          </option>
          <worldbody>
            <body>
              <joint type="free"/>
              <geom size="0.1"/>
            </body>
          </worldbody>
        </mujoco>
        """)
    d = mujoco.MjData(m)
    mujoco.mj_forward(m, d)
    mx = mjx.put_model(m)
    dx = mjx.put_data(m, d)

    dx = jax.jit(mjx.rne)(mx, dx)
    np.testing.assert_allclose(dx.qfrc_bias, 0)

if __name__ == '__main__':
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for testing."""

import sys
from typing import Dict, List, Tuple
from xml.etree import ElementTree as ET

from etils import epath
import mujoco
import numpy as np

TEST_FILES: List[str] = [
    'constraints.xml',
    'convex.xml',
    'pendula.xml',
]

_ACTUATOR_TYPES = ['motor', 'velocity', 'position', 'general', 'intvelocity']
_JOINT_TYPES = ['free', 'hinge', 'slide', 'ball']
_JOINT_AXES = ['1 0 0', '0 1 0', '0 0 1']
_FRICTIONS = ['1.2 0.003 0.0002', '0.2 0.0001 0.0005']
_KP_POS = ['1', '2']
_KP_INTVEL = ['10000', '2000']
_KV_VEL = ['123', '1']
_PAIR_FRICTIONS = ['1.2 0.9 0.003 0.0002 0.0001']
_SOLREFS = ['0.04 1.01', '0.05 1.02', '0.03 1.1', '0.015 1.0']
_SOLIMPS = [
    '0.75 0.94 0.002 0.2 2',
    '0.8 0.99 0.001 0.3 6',
    '0.6 0.9 0.003 0.1 1',
]
_DIMS = ['3']
_MARGINS = ['0.0', '0.01', '0.02']
_GAPS = ['0.0', '0.005']
_GEARS = ['20', '50', '100']


def p(pct: int) -> bool:
  assert 0 <= pct <= 100
  return np.random.uniform(low=0, high=100) < pct


def _make_joint(joint_type: str, name: str) -> Dict[str, str]:
  """Returns attributes for a joint."""
  joint_attr = {'type': joint_type, 'name': name}

  if joint_type not in ('free', 'ball'):
    joint_attr['axis'] = np.random.choice(_JOINT_AXES)
    lb, ub = -np.random.uniform() * 90, np.random.uniform() * 90
    joint_attr['range'] = f'{lb:.2f} {ub:.2f}'
  elif joint_type == 'ball':
    joint_attr['axis'] = '1 0 0'
    ub = np.random.uniform() * 90
    joint_attr['range'] = f'0.0 {ub:.2f}'

  if p(50) and joint_type != 'free':
    lb, ub = -np.random.uniform(), np.random.uniform()
    joint_attr['actuatorfrcrange'] = f'{lb:.2f} {ub:.2f}'

  if joint_type not in ('free',):
    joint_attr['damping'] = '{:.2f}'.format(np.random.uniform() * 20)
    joint_attr['stiffness'] = '{:.2f}'.format(np.random.uniform() * 20)

  return joint_attr


def _geom_solparams(
    pair: bool = False, enable_contact: bool = True
) -> Dict[str, str]:
  """Returns geom solver parameters."""
  params = {
      'contype': np.random.choice(['0', '1']) if enable_contact else '0',
      'conaffinity': np.random.choice(['0', '1']) if enable_contact else '0',
      'priority': np.random.choice(['-1', '2']),
      'solmix': np.random.choice(['0.0', '1.6']),
      'friction': np.random.choice(_FRICTIONS),
      'condim': np.random.choice(_DIMS),
  }
  pair_params = {
      'solreffriction': np.random.choice(_SOLREFS),
      'friction': np.random.choice(_PAIR_FRICTIONS),
      'condim': np.random.choice(_DIMS),
  }
  params = pair_params if pair else params
  params.update({
      'solimp': np.random.choice(_SOLIMPS),
      'solref': np.random.choice(_SOLREFS),
      'margin': np.random.choice(_MARGINS),
      'gap': np.random.choice(_GAPS),
  })

  return params


def _make_geom(
    pos: str, size: float, name: str, enable_contact: bool = True
) -> Dict[str, str]:
  """Returns attributes for a sphere geom."""
  attr = {
      'pos': pos,
      'type': 'sphere',
      'name': name,
      'size': f'{size:.2f}',
      'mass': '1',
  }
  attr.update(_geom_solparams(pair=False, enable_contact=enable_contact))

  return attr


def _make_actuator(actuator_type: str, joint: str) -> Dict[str, str]:
  """Returns attributes for an actuator."""
  attr = {'joint': joint}
  if actuator_type == 'motor':
    attr['gear'] = np.random.choice(_GEARS)
  elif actuator_type == 'position':
    attr['kp'] = np.random.choice(_KP_POS)
  elif actuator_type == 'general':
    attr['biastype'] = 'affine'
    attr['gainprm'] = '35 0 0'
    attr['biasprm'] = '0 -35 -0.65'
  elif actuator_type == 'intvelocity':
    attr['kp'] = np.random.choice(_KP_INTVEL)
    lb, ub = -np.random.uniform(), np.random.uniform()
    attr['actrange'] = f'{lb:.2f} {ub:.2f}'
  elif actuator_type == 'velocity':
    attr['kv'] = np.random.choice(_KV_VEL)

  if p(50) and actuator_type != 'intvelocity':
    lb, ub = -np.random.uniform(), np.random.uniform()
    attr['ctrlrange'] = f'{lb:.2f} {ub:.2f}'

  if p(50):
    lb, ub = -np.random.uniform(), np.random.uniform()
    attr['forcerange'] = f'{lb*10:.2f} {ub*10:.2f}'

  return attr


def create_mjcf(
    seed: int,
    min_trees: int = 1,
    max_trees: int = 1,
    max_tree_depth: int = 5,
    body_pos: Tuple[float, float, float] = (0.0, 0.0, -0.5),
    geom_pos: Tuple[float, float, float] = (0.0, 0.0, 0.0),
    max_stacked_joints=4,
    max_geoms_per_body=2,
    max_contact_excludes=1,
    max_contact_pairs=4,
    disable_actuation_pct: int = 0,
    add_actuators: bool = False,
    root_always_free: bool = False,
    enable_contact: bool = True,
) -> str:
  """Creates a random MJCF for testing.

  Args:
    seed: seed for rng
    min_trees: minimum number of kinematic trees to generate
    max_trees: maximum number of kinematic trees to generate
    max_tree_depth: the maximum tree depth
    body_pos: the default body position relative to the parent
    geom_pos: the default geom position in the body frame
    max_stacked_joints: maximum number of joints to stack for each body
    max_geoms_per_body: maximum number of geoms per body
    max_contact_excludes: maximum number of bodies to exlude from contact
    max_contact_pairs: maximum number of explicit geom contact pairs in the xml
    disable_actuation_pct: the percentage of time to disable actuation via the
      disable flag
    add_actuators: whether to add actuators
    root_always_free: if True, the root body of each kinematic tree has a free
      joint with the world
    enable_contact: if False, disables all contacts via contype/conaffinity

  Returns:
    an XML string for the MuJoCo config
  Raises:
    AssertionError when args are not in the correct ranges
  """
  np.random.seed(seed)

  assert min_trees <= max_trees
  assert max_tree_depth >= 1
  assert 0 <= disable_actuation_pct <= 100
  assert max_stacked_joints >= 1
  assert max_geoms_per_body >= 1
  assert max_contact_excludes >= 1
  assert max_contact_pairs >= 1

  mjcf = ET.Element('mujoco')
  opt = ET.SubElement(mjcf, 'option', {'timestep': '0.005', 'solver': 'CG'})
  world = ET.SubElement(mjcf, 'worldbody')
  ET.SubElement(mjcf, 'compiler', {'autolimits': 'true'})

  # disable flags
  if p(disable_actuation_pct):
    ET.SubElement(opt, 'flag', {'actuation': 'disable'})

  ET.SubElement(
      world,
      'geom',
      {
          'name': 'plane',
          'type': 'plane',
          'contype': '1' if enable_contact else '0',
          'conaffinity': '1' if enable_contact else '0',
          'size': '40 40 40',
      },
  )

  # kinematic trees
  tree_depth = np.random.randint(1, max_tree_depth + 1)

  def make_tree(body: ET.Element, depth: int) -> None:
    if depth >= tree_depth:
      return

    z_pos = np.random.uniform(low=-1, high=1) * 0.01  # small jitter
    pos = f'{body_pos[0]:.3f} {body_pos[1]:.3f} {body_pos[2] + z_pos:.3f}'
    n_bodies = len(list(mjcf.iter('body')))
    child = ET.SubElement(body, 'body', {'pos': pos, 'name': f'body{n_bodies}'})

    n_joints = len(list(mjcf.iter('joint')))
    for nj in range(np.random.randint(1, max_stacked_joints + 1)):
      joint_type = np.random.choice(_JOINT_TYPES)
      if nj == 0 and depth == 0 and root_always_free:
        joint_type = 'free'

      # free joint only allowed at top level
      while joint_type == 'free' and (depth > 0 or nj > 0):
        joint_type = np.random.choice(_JOINT_TYPES)

      joint_attr = _make_joint(joint_type, name=f'joint{n_joints + nj}')
      ET.SubElement(child, 'joint', joint_attr)

      prev_joints = child.findall('joint')
      had_ball_or_free = any(
          [j.get('type') in ('ball', 'free') for j in prev_joints]
      )
      if had_ball_or_free:
        break  # do not stack more joints

    n_geoms = len(list(mjcf.iter('geom')))
    for _ in range(np.random.randint(1, max_geoms_per_body + 1)):
      pos = ('{:.2f} ' * 3).format(*geom_pos).strip()
      size = 0.2 + np.random.uniform(low=-1, high=1) * 0.02
      geom_attr = _make_geom(
          pos, size, name=f'geom{n_geoms}', enable_contact=enable_contact
      )
      ET.SubElement(child, 'geom', geom_attr)
      n_geoms += 1

    make_tree(child, depth + 1)

  num_trees = np.random.randint(min_trees, max_trees + 1)
  for _ in range(num_trees):
    make_tree(world, 0)

  # actuators
  if add_actuators:
    actuator = ET.SubElement(mjcf, 'actuator')
    n_joints = len(list(mjcf.iter('joint')))
    nu = np.random.randint(1, n_joints + 1)
    actuators = []
    for i in range(nu):
      actuator_type = np.random.choice(_ACTUATOR_TYPES)
      attr = _make_actuator(actuator_type, joint=f'joint{i}')
      actuators.append((actuator_type, attr))

    np.random.shuffle(actuators)
    for typ, attr in actuators:
      ET.SubElement(actuator, typ, attr)

  # contact pairs
  contact = ET.SubElement(mjcf, 'contact')
  geoms = list(mjcf.iter('geom'))
  geom_names = [geom.get('name') for geom in geoms]
  n_geoms = len(geoms)
  pairs = set()
  for _ in range(min(max_contact_pairs, n_geoms * (n_geoms - 1) // 2)):
    if p(80):
      continue

    geom1, geom2 = np.random.choice(geom_names, replace=False, size=2)
    if geom1 > geom2:
      geom1, geom2 = geom2, geom1

    if (geom1, geom2) in pairs:
      continue

    pairs.add((geom1, geom2))
    attr = {'geom1': geom1, 'geom2': geom2}
    attr.update(_geom_solparams(pair=True))
    ET.SubElement(contact, 'pair', attr)

  # exclude contacts
  bodies = list(mjcf.iter('body'))
  body_names = [b.get('name') for b in bodies]
  n_bodies = len(bodies)
  for _ in range(min(max_contact_excludes, (n_bodies * (n_bodies - 1) // 2))):
    if p(50):
      continue

    body1, body2 = np.random.choice(body_names, replace=False, size=2)
    ET.SubElement(contact, 'exclude', {'body1': body1, 'body2': body2})

  # ElementTree.indent is not available before Python 3.9
  if sys.version_info.minor >= 9:
    ET.indent(mjcf)

  return ET.tostring(mjcf).decode('utf-8')


def load_test_file(name: str) -> mujoco.MjModel:
  """Loads a mujoco.MjModel based on the file name."""
  path = epath.resource_path('mujoco.mjx') / 'test_data' / name
  m = mujoco.MjModel.from_xml_path(path.as_posix())
  return m
